/*  This file is part of distributed Structured Prediction (dSP) - http://www.alexander-schwing.de/
*
*  distributed Structured Prediction (dSP) is free software: you can
*  redistribute it and/or modify it under the terms of the GNU General
*  Public License as published by the Free Software Foundation, either
*  version 3 of the License, or (at your option) any later version.
*
*  distributed Structured Prediction (dSP) is distributed in the hope
*  that it will be useful, but WITHOUT ANY WARRANTY; without even the
*  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
*  PURPOSE. See the GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with distributed Structured Prediction (dSP).
*  If not, see <http://www.gnu.org/licenses/>.
*
*  Copyright (C) 2010-2013  Alexander G. Schwing  [http://www.alexander-schwing.de/]
*/

//Author: Alexander G. Schwing

/*  ON A PERSONAL NOTE: I spent a significant amount of time to go through
*  both, theoretical justifications and coding of this framework.
*  I hope the package is useful for your task. Any requests,
*  feedback, donations and support would be greatly appreciated.
*  Thank you for contacting me!
*/

#ifndef __LIB_CRF__
#define __LIB_CRF__

#include "../Settings.h"

#include <set>
#include <map>
#include <vector>
#include <stddef.h>

template <class T>
struct CRFSettings {
	T C;
	T epsilon;
	T p;
	T rG;
	size_t CRFIterations;
	CRFSettings() : C(T(1.0)), epsilon(T(1.0)), p(T(2.0)), rG(T(1e-5)), CRFIterations(100) {}
	CRFSettings(T C, T epsilon, T p, T relGap, size_t CRFIterations) : C(C), epsilon(epsilon), p(p), rG(relGap), CRFIterations(CRFIterations) {}
};

template <class T>
struct PotentialStruct {
	T* pot;
	size_t* num_states;
	size_t num_variables;
	PotentialStruct() : pot(NULL), num_states(NULL), num_variables(-1) {;}
	PotentialStruct(T* p, size_t* num_s, size_t num_v) : pot(p), num_states(num_s), num_variables(num_v) {;}
	float AccessVariable(const size_t* current_state) const {
		size_t retVal = current_state[num_variables-1];
		for(size_t k=num_variables-2;k>=0;--k) {
			retVal = retVal*num_states[k+1] + current_state[k];
		}
		return pot[retVal];
	}
};

template <class T>
struct InternalNodeData {
	bool hidden;
	size_t NumStates;
	struct PotentialStruct<T>* NodeLoss;
	std::map<size_t, struct PotentialStruct<T>* > phi_r;
	std::map<size_t, std::vector<T>* > ConnectedTo;		//second is a pointer to the message
	InternalNodeData() : NodeLoss(NULL) {};
};

template <class T>
struct InternalFactorData {
	struct PotentialStruct<T>* FactorLoss;
	std::map<size_t, struct PotentialStruct<T>*> phi_r;
	std::map<size_t, std::vector<T>*> lambda_v;
	InternalFactorData() : FactorLoss(NULL) {};
};

template <class T>
struct VertexData {
	struct PotentialStruct<T>* potData;
	bool hidden;
	size_t NumStates;
	std::map<size_t, struct PotentialStruct<T>*> ConnectedToFactor;
	VertexData() : potData(NULL), hidden(false) {};
	VertexData(struct PotentialStruct<T>* p, size_t NumStat, bool hid) : potData(p), hidden(hid), NumStates(NumStat) {};
	VertexData(std::map<size_t, struct PotentialStruct<T>* > &CTF, size_t NumStat, bool hid) : potData(NULL), NumStates(NumStat), hidden(hid) {ConnectedToFactor = CTF; }
	VertexData(struct PotentialStruct<T>* p, std::map<size_t, struct PotentialStruct<T>* > &CTF, size_t NumStat, bool hid) : potData(p), NumStates(NumStat), hidden(hid) {ConnectedToFactor = CTF; }
};

template <class T>
struct Feature2 {
	std::vector<std::map<size_t, struct VertexData<T> > > Vertices;		//first index is training set x, then node/vertex index
};

template <class T>
class libCRF {
private:
	T epsilon;
	T p;
	T C;
	T rG;
	size_t max_iterations;

	std::vector<struct Feature2<T> > *FeatureArray;
	std::vector<std::vector<std::map<size_t, InternalFactorData<T>* > > > EdgeSet;

	std::vector<std::map<size_t, InternalNodeData<T> > > Vertices;
	std::vector<std::map<size_t, InternalFactorData<T> > > Factors;		//the second map, second element is a pointer to the message from this node to the factor
	std::vector<std::map<size_t, std::map<size_t, std::vector<T> > > > Lambda;

	int computeFactorBelief(size_t x, InternalFactorData<T>* alpha, std::vector<T> *factorBelief_x_alpha, std::vector<T> &theta);
	int computeLocalBelief(size_t x, size_t v_hat, std::vector<T> *locBel_x_v, std::vector<T> &theta);
	int computeBeliefs(std::vector<std::map<size_t, std::vector<T> > > &localBelief, std::vector<std::map<size_t, std::vector<T> > > &factorBelief, std::vector<T> &theta);
	T computePrimal(std::vector<T> &theta, std::vector<T> &empiricalMeans);
	T computeDual(std::vector<T> &theta, std::vector<T> &empiricalMeans, bool outputBeliefs);
	T computePrimalOfFeature(size_t cur_r, std::vector<T> &theta, std::vector<T> &empiricalMeans);
	
	int logNorm(std::vector<T> &FinalResult, const std::vector<T> &Input, size_t DimOfInterest, size_t numVars, const size_t* const num_states, const size_t* const stateProducts);
	T logNorm(std::vector<T> &Input);
	T entropy(std::vector<T> &Input);

	int WriteUAIInferenceFile(const char* fn, std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > > &LocalPot, std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > > &FactorPot);
public:
	libCRF();
	~libCRF();

	void SetParameters(struct CRFSettings<T>& params) {C = params.C;epsilon = params.epsilon;p = params.p;rG = params.rG;max_iterations=params.CRFIterations;}

	int Initialize(std::vector<struct Feature2<T> > &FeatureArray, struct Feature2<T>* Loss, size_t TotalTrainExamples);
	T Iterate(std::vector<T> *theta, std::vector<T> &empiricalMean);
	int ResetMessages();
	double ComputeMAP(std::vector<T> *theta, std::vector<std::map<size_t, std::vector<T> > >& AllLocalBeliefs, const char* folder);
	int Clear();
};

#endif