/*  This file is part of distributed Structured Prediction (dSP) - http://www.alexander-schwing.de/
*
*  distributed Structured Prediction (dSP) is free software: you can
*  redistribute it and/or modify it under the terms of the GNU General
*  Public License as published by the Free Software Foundation, either
*  version 3 of the License, or (at your option) any later version.
*
*  distributed Structured Prediction (dSP) is distributed in the hope
*  that it will be useful, but WITHOUT ANY WARRANTY; without even the
*  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
*  PURPOSE. See the GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with distributed Structured Prediction (dSP).
*  If not, see <http://www.gnu.org/licenses/>.
*
*  Copyright (C) 2010-2013  Alexander G. Schwing  [http://www.alexander-schwing.de/]
*/

//Author: Alexander G. Schwing

/*  ON A PERSONAL NOTE: I spent a significant amount of time to go through
*  both, theoretical justifications and coding of this framework.
*  I hope the package is useful for your task. Any requests,
*  feedback, donations and support would be greatly appreciated.
*  Thank you for contacting me!
*/

#include "mex.h"
#include "matrix.h"
#include "assert.h"
#include <string.h>
#include <stddef.h>

#include "../libCRF/libCRF.h"

template<class T>
int ProcessElement(const mxArray *element_r, struct Feature2<T> &f_r) {
	const mxArray *sample = mxGetField(element_r, 0, "sample");
	if(sample==NULL) {
		return -1;
	}

	size_t train_r = mxGetNumberOfElements(sample);

	for(size_t x=0;x<train_r;++x) {
		std::map<size_t, struct VertexData<T> > NodeData;	//<Node ID; VertexData>

		const mxArray *sample_x = mxGetCell(sample, x);
		if(sample_x==NULL) {
			f_r.Vertices.push_back(NodeData);
			continue;
		}

		const mxArray *locFeature = mxGetField(sample_x, 0, "local");
		if(locFeature==NULL) {
			f_r.Vertices.push_back(NodeData);
			continue;
		}

		const mxArray *facFeature = mxGetField(sample_x, 0, "factor");
		if(facFeature==NULL) {
			f_r.Vertices.push_back(NodeData);
			continue;
		}

		size_t Nv = mxGetNumberOfElements(locFeature);

		for(size_t v=0;v<Nv;++v) {
			const mxArray *locFeature_v = mxGetCell(locFeature, v);
			if(locFeature_v==NULL) {
				continue;
			}

			const mxArray *NumStates_v = mxGetField(locFeature_v, 0, "NumStates");
			if(NumStates_v==NULL) {
				continue;
			}
			size_t NumStates = size_t(*mxGetPr(NumStates_v));

			const mxArray *pot_v = mxGetField(locFeature_v, 0, "pot");
			if(pot_v!=NULL) {
				size_t v_sz = mxGetNumberOfElements(pot_v);
				
				assert(NumStates==v_sz);

				NodeData[v].hidden = false;
				NodeData[v].NumStates = NumStates;
				double* v_Pot = mxGetPr(pot_v);
				NodeData[v].potData = new struct PotentialStruct<T>(v_Pot, &NodeData[v].NumStates, 1);
			}

			const mxArray *connTo_v = mxGetField(locFeature_v, 0, "connTo");
			if(connTo_v!=NULL) {
				size_t alpha_sz = mxGetNumberOfElements(connTo_v);
				double* alpha_vec = mxGetPr(connTo_v);

				NodeData[v].NumStates = NumStates;

				for(size_t alpha=0;alpha<alpha_sz;++alpha) {
					const mxArray *facFeature_alpha = mxGetCell(facFeature, size_t(alpha_vec[alpha])-1 );
					if(facFeature_alpha==NULL) {
						continue;
					}

					const mxArray *pot_alpha = mxGetField(facFeature_alpha, 0, "pot");
					const mxArray *sz_alpha = mxGetField(facFeature_alpha, 0, "size");
					if(pot_alpha!=NULL && sz_alpha!=NULL) {
						double* alpha_Pot = mxGetPr(pot_alpha);

						size_t alpha_sz = mxGetNumberOfElements(sz_alpha);
						double* alpha_SZVec = mxGetPr(sz_alpha);
						size_t* facSZ = new size_t[alpha_sz];
						for(size_t k=0;k<alpha_sz;++k) {
							facSZ[k] = size_t(alpha_SZVec[k]);
						}

						NodeData[v].ConnectedToFactor[size_t(alpha_vec[alpha])] = new struct PotentialStruct<T>(alpha_Pot, facSZ, alpha_sz);
					}
				}
			}
		}

		f_r.Vertices.push_back(NodeData);
	}

	return 0;
}

void ClearStructure(struct Feature2<double> &tmp) {
	for(size_t x=0;x<tmp.Vertices.size();++x) {
		for(std::map<size_t, struct VertexData<double> >::iterator v=tmp.Vertices[x].begin();v!=tmp.Vertices[x].end();++v) {
			for(std::map<size_t, struct PotentialStruct<double>*>::iterator alpha=v->second.ConnectedToFactor.begin();alpha!=v->second.ConnectedToFactor.end();++alpha) {
				delete [] alpha->second->num_states;
				delete alpha->second;
			}

			delete v->second.potData;
		}
	}
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

	size_t sz = mxGetNumberOfElements(prhs[0]);

	std::vector<struct Feature2<double> > FeatureArray(sz);
	struct Feature2<double> Loss;

	for(size_t r=0;r<sz;++r) {
		const mxArray *feature_r = mxGetCell(prhs[0], r);
		if(feature_r==NULL) {
			continue;
		}
		
		ProcessElement<double>(feature_r, FeatureArray[r]);
	}

	int WithLoss = ProcessElement<double>(prhs[1], Loss);

	struct CRFSettings<double> params;
	std::vector<double> theta(FeatureArray.size(), 1.0);

	double* theta_init = mxGetPr(prhs[2]);
	std::copy(theta_init, theta_init+theta.size(), &theta[0]);

	params.epsilon = *mxGetPr(prhs[3]);

	double what = *mxGetPr(prhs[4]);

	if(what==1) { //train
		double* empiricalMeans_Arr = mxGetPr(prhs[5]);
		std::vector<double> empiricalMeans(empiricalMeans_Arr, empiricalMeans_Arr+sz);

		if(nrhs>6)
			params.p = *mxGetPr(prhs[6]);
		if(nrhs>7)
			params.C = *mxGetPr(prhs[7]);
		if(nrhs>8)
			params.rG = *mxGetPr(prhs[8]);
		if(nrhs>9)
			params.CRFIterations = size_t(*mxGetPr(prhs[9]));

		libCRF<double> CRF;

		CRF.SetParameters(params);

		if(WithLoss==-1) {
			CRF.Initialize(FeatureArray, NULL, FeatureArray[0].Vertices.size());
		} else {
			CRF.Initialize(FeatureArray, &Loss, FeatureArray[0].Vertices.size());
		}

		CRF.Iterate(&theta, empiricalMeans);

		plhs[0] = mxCreateDoubleMatrix(sz, 1, mxREAL);
		double* output = mxGetPr(plhs[0]);

		memcpy(output, &theta[0], sz*sizeof(double));
	} else {
		char* folder = NULL;
		if(nrhs>10 && mxIsChar(prhs[10])==1 && mxGetM(prhs[10])==1) {
			folder = mxArrayToString(prhs[10]);
		}

		libCRF<double> CRF;

		CRF.SetParameters(params);

		CRF.Initialize(FeatureArray, NULL, FeatureArray[0].Vertices.size());

		std::vector<std::map<size_t, std::vector<double> > > AllLocalBeliefs;

		CRF.ComputeMAP(&theta, AllLocalBeliefs, folder);

		if(nlhs>0) {
			mwSize dims = AllLocalBeliefs.size();
			const char *fName[] = {"node"};
			plhs[0] = mxCreateStructArray(1, &dims, 1, fName);

			for(std::vector<std::map<size_t, std::vector<double> > >::iterator x=AllLocalBeliefs.begin();x!=AllLocalBeliefs.end();++x) {
				size_t maxVal = x->rbegin()->first + 1;

				mxArray* x_out = mxCreateCellArray(1, &maxVal);

				for(std::map<size_t, std::vector<double> >::iterator v=x->begin();v!=x->end();++v) {
					mxArray* tmp = mxCreateDoubleMatrix(v->second.size(), 1, mxREAL);
					double* output = mxGetPr(tmp);
					memcpy(output, &v->second[0], v->second.size()*sizeof(double));

					mxSetCell(x_out, v->first, tmp);
				}

				mxSetField(plhs[0], x-AllLocalBeliefs.begin(), "node", x_out);
			}
		}
	}

	for(size_t r=0;r<FeatureArray.size();++r) {
		ClearStructure(FeatureArray[r]);
	}
	ClearStructure(Loss);
}